"""
This script runs the reinforcement learner on a selected app.
"""

import os
import subprocess

from typing import List, Dict
import logging
import json
import traceback
from flask import current_app

from RLTester.utils.progress_tracker import ProgressTracker

from RLTester.apk_util import extract_info_from_apk, is_apk
from RLTester.adapt_textproto import update_textproto

logger = logging.getLogger(__name__)

def execute_tests(test_iteration: str, storage_path: str, progress_callback=None) -> List[Dict]:
    """
    Execute all tests for a given iteration with comprehensive reporting.
    
    Args:
        test_iteration (str): Test iteration identifier
        storage_path (str): Base storage path for test files
        progress_callback (callable): Callback function for progress updates
        
    Returns:
        List[Dict]: Final list of test results
    """
    logger.info(f"Starting test execution for iteration {test_iteration}")
    results = []
    
    # Initialize progress tracker
    progress = ProgressTracker()
    if progress_callback:
        progress.set_progress_callback(progress_callback)

    try:
        # Initial progress update
        progress.emit_progress(
            'init',
            f'Starting execution'
        )

        

        # Final progress update
        progress.emit_progress(
            'complete',
            f'Completed execution'
        )
    except Exception as e:
        logger.error(f"Error executing tests: {str(e)}")
        logger.error(traceback.format_exc())
        progress.emit_progress('error')
    

def run():

    storage_path = current_app.config['STORAGE_PATH']
    app_path = os.path.join(storage_path, 'APK.apk')
    result_path = os.path.join(storage_path, 'RL_Results')

    # Check if the given path is an apk file
    if not is_apk(app_path):
        raise ValueError("The given file is not an apk file")

    package_name, main_activity = extract_info_from_apk(app_path)
    update_textproto(app_path, package_name, main_activity)

    # Command to run the reinforcement learner with the adapted textproto file
    command = [
        'python3', '/home/coder/workspace/Testplattform/ReinforcementTest/AppTestingEnvironment/reinforced_testing/android_training.py',
        '--avd_name=avd1',
        '--android_avd_home=/home/coder/.android/avd/',
        '--android_sdk_root=/opt/android-sdk',
        '--emulator_path=/opt/android-sdk/emulator/emulator',
        '--adb_path=/opt/android-sdk/platform-tools/adb',
        '--num_steps=500',
        '--task_path=/home/coder/workspace/Testplattform/ReinforcementTest/custom_app.textproto',
        '--run_headless=true',
        '--rl_output_widget_path=' + result_path
    ]

    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print("Error running random agent:", e)
