"""
Responsible for a graphical representation (as an interactive coordinate system) of
key-value pairs from the output of the reinforcement learner.
"""
import json
import os

import plotly.graph_objects as go
from plotly.subplots import make_subplots


DEFAULT_SIZE = 1


def create_graph(name, key, value):
    """
    Creates a Plotly graph with the given name, key (x-axis), and value (y-axis).

    Parameters:
    - name (str): Title of the graph.
    - key (str): Label for the x-axis.
    - value (str): Label for the y-axis.

    Returns:
    - fig: Plotly Figure object.
    """

    fig = make_subplots(rows=DEFAULT_SIZE, cols=DEFAULT_SIZE)

    fig.update_layout(
        title=name,
        xaxis=dict(title=key),
        yaxis=dict(title=value),
    )

    return fig


def add_step_reward(fig, key, value):
    """
    Adds a step reward data point to an existing Plotly graph.

    Parameters:
    - fig: Plotly Figure object to which the data point will be added.
    - key: X-axis value of the data point.
    - value: Y-axis value of the data point.

    Note:
    This function modifies the provided 'fig' object in-place.
    """
    fig.add_trace(go.Scatter(x=[key], y=[value], mode='markers', marker=dict(size=5, color='red')))


def extract_graph(fig, path, performance):
    """
    Saves the JSON representation of a Plotly graph and its details to a JSON file.

    Parameters
        fig: Plotly Figure object to be converted to JSON.
        path: Path where JSON file gets stored.
        performance: Performance data to be saved along with the graph.
        
    Returns:
        None
    """
    print("PATH: " + path)
    if not os.path.exists(path):
        os.makedirs(path)

    files = os.listdir(path)
    print("LISTING FILES: " + str(files))
    if files:
        highest_iteration = max([int(filename.split('_')[0]) for filename in files])
    else:
        highest_iteration = 0

    filename = f"{highest_iteration + 1}_rl_result.json"
    file_path = os.path.join(path, filename)

    fig_json = fig.to_json()

    result_data = {"performance": performance, "graph": fig_json}

    with open(file_path, 'w') as json_file:
        json.dump(result_data, json_file)
