# progress tracker for rltester

import logging
from typing import Optional, Callable, Dict, Any
logger = logging.getLogger(__name__)

class ProgressTracker:
    """
    Handles progress tracking and callback management for the test generation process.
    
    Attributes:
        total_files (int): Total number of files to process
        total_methods (int): Total number of methods to process
        current_file (int): Current file being processed
        current_method (int): Current method being processed
        progress_callback (Optional[Callable]): Callback function for progress updates
    """
    
    def __init__(self):
        """Initialize the progress tracker with default values."""
        self.total_files: int = 0
        self.total_methods: int = 0
        self.current_file: int = 0
        self.current_method: int = 0
        self.progress_callback: Optional[Callable] = None
        logger.info("Progress tracker initialized")

    def set_progress_callback(self, callback: Callable) -> None:
        """
        Set callback function for progress updates.
        
        Args:
            callback: Function to be called with progress updates
        """
        self.progress_callback = callback
        logger.debug("Progress callback set")

    def emit_progress(self, status: str, message: str, additional_data: Optional[Dict[str, Any]] = None) -> None:
        """
        Emit progress update through callback if set.
        
        Args:
            status: Current status of the operation
            message: Progress message
            additional_data: Optional additional data to include in progress update
        """
        if self.progress_callback:
            try:
                # Estimate total methods if not set
                if self.total_methods == 0 and status in ['processing', 'generating']:
                    self.total_methods = max(10, self.total_methods * 2)
                
                progress = {
                    'status': status,
                    'message': message,
                    'current_file': self.current_file,
                    'total_files': max(1, self.total_files),
                    'current_method': self.current_method,
                    'total_methods': max(1, self.total_methods)
                }
                
                # Calculate percentage
                if self.total_methods > 0:
                    progress['percent'] = int((self.current_method / self.total_methods) * 100)
                else:
                    progress['percent'] = 0
                
                if additional_data:
                    progress.update(additional_data)
                
                self.progress_callback(progress)
            except Exception as e:
                logger.error(f"Failed to emit progress update: {e}")