from stable_baselines3.common.callbacks import BaseCallback
from datetime import datetime

class RLProgressCallback(BaseCallback):
    """
    Custom callback for reporting training progress from stable baselines 3.
    Uses standardized status types matching the frontend expectations.
    """
    def __init__(self, progress_callback, mode='train', verbose=0):
        super().__init__(verbose)
        self.progress_callback = progress_callback
        self.mode = mode
        self.errors_found = []  # For tracking errors during evaluation
        
    def _on_training_start(self):
        """Called when training starts."""
        message = 'Initializing RL training...' if self.mode == 'train' else 'Starting error search...'
        self.progress_callback({
            'status': 'init',
            'message': message,
            'timestamp': datetime.now().strftime('%H:%M:%S'),
            'percent': 0
        })

    def _on_step(self) -> bool:
        """Called after each training step."""
        # Calculate progress percentage
        progress = (self.num_timesteps / self.locals['total_timesteps']) * 100
        
        # Get the episode reward if available
        episode_reward = 0
        if len(self.locals.get('episode_rewards', [])) > 0:
            episode_reward = self.locals['episode_rewards'][-1]
        
        if self.mode == 'train':
            message = f'Training step {self.num_timesteps}/{self.locals["total_timesteps"]} - Reward: {episode_reward:.2f}'
        else:
            message = f'Searching step {self.num_timesteps}/{self.locals["total_timesteps"]} - Found errors: {len(self.errors_found)}'
            
        self.progress_callback({
            'status': 'processing',
            'message': message,
            'timestamp': datetime.now().strftime('%H:%M:%S'),
            'percent': round(progress, 1),
            'reward': episode_reward,
            'timestep': self.num_timesteps,
            'errors_found': len(self.errors_found) if self.mode == 'evaluate' else None
        })
        return True
        
    def _on_training_end(self):
        """Called when training/evaluation ends."""
        if self.mode == 'train':
            message = 'RL training completed successfully'
        else:
            message = f'Error search completed. Found {len(self.errors_found)} potential issues'
            
        self.progress_callback({
            'status': 'complete',
            'message': message,
            'timestamp': datetime.now().strftime('%H:%M:%S'),
            'percent': 100,
            'errors_found': self.errors_found if self.mode == 'evaluate' else None
        })

    def log_error(self, error_info):
        """Log an error found during evaluation."""
        if self.mode == 'evaluate':
            self.errors_found.append(error_info) 
