"""
A module to train an SAC (Soft Actor-Critic) agent using the Android environment.

This module initializes the Android environment with given settings, runs the SAC agent
through a specified number of steps, and optionally logs results with Weights & Biases (wandb).

Key Features:
- Customizable run with various parameters.
- Converts action format to match the environment's expectations.
- Inefficient by design to minimize external dependencies.

"""

import importlib
import sys
import time

import numpy as np
import torch
from absl import app
from absl import logging

# Add sys.path modification close to the sys import
sys.path.append('/home/coder/workspace/apptestingenvironment/reinforced_training')

import android_env.loader as loader
from calculator import reward_calculator
import wandb

# Global variables for training parameters
variables = {
    "num_steps": 1000,  # Total number of training steps
    "batch_size": 8,  # Batch size for training updates
    "action_dim": 3,  # Number of action dimensions
    "action_options": 3,  # Number of discrete action options
    "reset_after": 256,  # Steps before environment resets
    "training_agent_name": "sac_training_agent_2"  # SAC agent module name
}

def initialize(environment, reward_type) -> None:
    """
    Initializes the training environment and SAC agent module.
    
    Args:
        environment: The Android environment instance.
        reward_type: The reward calculation type to be used.
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    training_agent_name = variables["training_agent_name"]
    training_agent = importlib.import_module(training_agent_name)
    
    variables.update({
        "training_agent": training_agent,
        "device": device,
        "environment": environment,
        "reward_calculation": reward_type
    })

def set_run_parameters():
    """
    Define hyperparameters for training.
    This function is a placeholder for setting additional run parameters.
    """
    pass

def main(do_wandb):
    """
    Runs the SAC agent training loop in the Android environment.

    Args:
        do_wandb (bool): Whether to enable Weights & Biases (wandb) logging.
    """
    reward_type = variables["reward_calculation"]
    number_of_steps = variables["num_steps"]
    reward_calc = reward_calculator.RewardCalculator(reward_type)
    env = variables["environment"]

    timestep = env.reset()
    observation_space = torch.tensor(timestep.observation['pixels']).permute(2, 0, 1).numpy()
    input_shape = observation_space.shape
    
    training_agent = variables["training_agent"]
    device = variables["device"]
    agent = training_agent.SACAgent(input_shape, variables["action_dim"], 256, device)
    
    prev_observation = (torch.tensor(timestep.observation['pixels']).permute(2, 0, 1).numpy(),
                        torch.zeros(variables["action_dim"]))
    done = True
    reward_sum = 0
    
    for step in range(number_of_steps):
        if done:
            print("RESET")
            if do_wandb:
                wandb.log({"reward_end": reward_sum})
            reward_sum = 0
            reward_calc.reset()
            env.reset()
            prev_observation = get_observation(reward_calc, 0, torch.tensor([0, 0, 0]), step, timestep, variables["reset_after"], variables["action_dim"], env)
        
        action, action_array, action_random = get_action(variables["action_dim"], variables["action_options"], agent, prev_observation)
        action_tensor = torch.tensor(action_array, dtype=torch.float32)
        timestep = execute_action(env, action, time.time(), 0.2)
        
        observation, done, reward_sum = get_observation(reward_calc, reward_sum, action_tensor, step, timestep, variables["reset_after"], variables["action_dim"], env)
        reward = calculate_reward(reward_calc, observation, prev_observation, done, variables["reset_after"], reward_type, action_tensor, torch.tensor([0, 0, 0]), reward_sum)
        
        save(agent, (prev_observation, action, action_tensor, observation, torch.tensor(done), reward))
        prev_observation = observation
        agent.train(variables["batch_size"])
        log_wandb(do_wandb, (prev_observation, action, action_tensor, observation, torch.tensor(done), reward), reward_sum)

def get_action(action_dim, action_options, agent, observation):
    """
    Selects and formats an action from the SAC agent.
    
    Args:
        action_dim: Number of action dimensions.
        action_options: Number of action options.
        agent: The SAC agent instance.
        observation: Current environment observation.
    
    Returns:
        Tuple of formatted action, action array, and boolean indicating randomness.
    """
    agent_action, was_random = agent.act(observation)
    return format_action_for_env(agent_action, action_dim, action_options), agent_action, was_random

def execute_action(env, action, start_time, wait_after_action):
    """
    Executes an action in the environment and manages timing constraints.
    
    Args:
        env: The environment instance.
        action: The action to execute.
        start_time: Time when action execution started.
        wait_after_action: Time delay after action execution.
    
    Returns:
        The timestep result after executing the action.
    """
    timestep = env.step(action)
    elapsed_time = time.time() - start_time
    time.sleep(max(0, wait_after_action - elapsed_time))
    return timestep

def calculate_reward(reward_calc, observation, prev_observation, done, reset_after, reward_type, action_tensor, prev_action_tensor, reward_sum):
    """
    Computes the reward using the reward calculator.
    
    Args:
        reward_calc: The reward calculator instance.
        observation: Current observation.
        prev_observation: Previous observation.
        done: Whether the episode is done.
        reset_after: Number of steps before reset.
        reward_type: Type of reward function.
        action_tensor: The current action tensor.
        prev_action_tensor: The previous action tensor.
        reward_sum: Accumulated reward sum.
    
    Returns:
        Computed reward value.
    """
    return reward_calc.calculate_reward_temp(observation, prev_observation)

def format_action_for_env(action, action_dim, options):
    """
    Converts raw SAC agent action to a format suitable for the Android environment.
    
    Args:
        action: Raw action output from the SAC agent.
        action_dim: Number of action dimensions.
        options: Number of action options.
    
    Returns:
        Dictionary representing the formatted action and its array.
    """
    action[0] = int(action[0] * action_dim)
    return {"action_type": action[0], "touch_position": action[1:]}, action

"""if __name__ == '__main__':
    logging.set_verbosity('info')
    app.run(main)
"""