import torch
import torch.nn as nn
import torchvision.models as models
import torchvision.transforms as transforms
import faiss
import numpy as np
from PIL import Image
import logging
import cv2

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class ImageObservationChecker:
    """
    Detects previously seen images using CNN-based feature extraction and FAISS indexing.

    Features:
    - Uses MobileNetV2 for lightweight image embeddings.
    - Stores embeddings in a FAISS index for fast similarity searches.
    - Supports GPU acceleration for both FAISS and CNN processing.
    """
    
    def __init__(self, image_size=(84, 84), threshold=0, use_gpu=True):
        """
        Initializes the feature extractor and FAISS index.

        Args:
            image_size (tuple): Target image dimensions (width, height).
            threshold (float): Distance threshold for detecting seen images.
            use_gpu (bool): Enables GPU usage for model inference and FAISS.
        """
        self.image_size = image_size
        self.threshold = threshold
        self.device = torch.device("cuda" if torch.cuda.is_available() and use_gpu else "cpu")

        # Load a pre-trained model (MobileNetV2 for lightweight embeddings)
        self.model = nn.Sequential(
            models.mobilenet_v2(pretrained=True).features,  # Feature extractor layers
            nn.AdaptiveAvgPool2d((1, 1))  # Global Average Pooling to 1x1x1280
        )
        self.model.eval()  # Set model to evaluation mode
        self.model.to(self.device)  # Move model to GPU if available

        # Set embedding size based on MobileNetV2 output
        self.vector_size = 1280

        # Initialize FAISS index
        self.index = self._initialize_faiss_index(use_gpu)

        # Define preprocessing transformations
        self.preprocess = transforms.Compose([
            transforms.Resize(self.image_size),  # Resize to the defined image size
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485], std=[0.229]),  # Grayscale normalization
        ])

        self.reference_embedding = None  # Store the first image's embedding in the current episode

    def _initialize_faiss_index(self, use_gpu):
        """
        Initializes a FAISS index with GPU support if available.

        :param use_gpu: If True, uses GPU for FAISS.
        :return: FAISS index object.
        """
        index = faiss.IndexFlatL2(self.vector_size)  # L2 (Euclidean) distance
        if use_gpu and faiss.get_num_gpus() > 0:
            logger.info("Using FAISS GPU index.")
            gpu_index = faiss.index_cpu_to_all_gpus(index)
            return gpu_index
        logger.info("Using FAISS CPU index.")
        return index

    def preprocess_observation(self, observation):
        """
        Converts input images to a normalized tensor format for CNN processing.

        Args:
            observation (numpy.ndarray): Input grayscale or RGB image.
        
        Returns:
            torch.Tensor: Processed image tensor.
        """

        if isinstance(observation, np.ndarray):
            if observation.ndim == 3 and observation.shape[-1] == 1:
                observation = observation.squeeze(-1)  # Convert (H, W, 1) to (H, W)
            if observation.ndim == 2:  # Grayscale to RGB
                observation = np.stack([observation] * 3, axis=-1)
            elif observation.ndim == 3 and observation.shape[-1] != 3:
                raise ValueError(f"[ERROR] Unsupported observation format: {observation.shape}")
            pil_image = Image.fromarray(observation.astype(np.uint8))
            return self.preprocess(pil_image)
        else:
            raise ValueError(f"[ERROR] Observation must be a numpy array, got {type(observation)}")

    def _preprocess_image_batch(self, images):
        """
        Preprocesses a batch of images into embeddings.

        :param images: List of input images (numpy arrays or PIL Images).
        :return: Batch of flattened feature vectors.
        """
        if isinstance(images, list):
            processed_images = [self.preprocess_observation(img) for img in images]
        else:
            processed_images = [self.preprocess_observation(images)]

        # Stack tensors and move to device
        image_batch = torch.stack(processed_images).to(self.device)

        # Extract embeddings
        with torch.no_grad():
            embeddings = self.model(image_batch).view(len(processed_images), -1).cpu().numpy()
        return embeddings.astype(np.float32)

    def add_image(self, image):
        """
        Adds an image embedding to the FAISS index.

        Args:
            image (numpy.ndarray or PIL.Image): Input image.
        """
        image_vector = self._preprocess_image_batch(image)[0]
        if self.index.ntotal == 0:
            self.reference_embedding = image_vector
        self.index.add(np.array([image_vector]))
        # logger.info(f"Image added to FAISS index. Current index size: {self.index.ntotal}")

    def calculate_smallest_distance(self, image):
        """
        Computes the smallest FAISS distance between the given image and stored embeddings.

        Args:
            image (numpy.ndarray or PIL.Image): Input image.

        Returns:
            float: Distance to the closest stored image.
        """
        
        if self.index.ntotal == 0:
            return float('inf')  # No images to compare against
        image_vector = self._preprocess_image_batch(image)[0]
        distances, _ = self.index.search(np.array([image_vector]), 1)  # Query for the nearest neighbor
        return distances[0][0]

    def is_seen(self, image):
        """
        Determines if an image has been seen before based on FAISS similarity.

        Args:
            image (numpy.ndarray or PIL.Image): Input image.

        Returns:
            tuple: (bool indicating if seen, smallest distance to stored images).
        """
        if self.index.ntotal == 0:
            self.add_image(image)
            return False, 0

        smallest_distance = self.calculate_smallest_distance(image)

        if smallest_distance > self.threshold:
            # logger.info(f"New image detected. Distance: {smallest_distance:.4f}. Adding to FAISS index.")
            self.add_image(image)
            return False, smallest_distance
        # logger.info(f"Image already seen. Distance: {smallest_distance:.4f}")
        return True, smallest_distance

    def get_reference_embedding(self):
        """
        Retrieves the first stored image embedding from the current session.

        Returns:
            numpy.ndarray or None: Reference embedding if available.
        """
        return self.reference_embedding

    def clear(self):
        """
        Clears all stored embeddings and resets the FAISS index.
        """
        self.index.reset()  # Reset FAISS index
        self.reference_embedding = None  # Clear reference embedding
        # logger.info("FAISS index cleared.")
