from android_env import loader
from android_env.components import config_classes
import wandb
import os
import subprocess
import wandb
from dotenv import load_dotenv
import torch
from stable_baselines3.common.env_checker import check_env
from wrapper import element_detection_reward_wrapper
from types import SimpleNamespace

load_dotenv()  # Load environment variables from .env file

def check_avd_exists(avd_name):
    """Check if the specified AVD name exists in the AVD manager list."""
    
    cmd = f"{os.path.join(os.getenv('ANDROID_SDK_ROOT', '/opt/android-sdk'), 'emulator', 'emulator')} -list-avds"
    result = subprocess.run(cmd, shell=True, capture_output=True, text=True)
    avds = result.stdout.splitlines()
    return avd_name in avds

def create_avds(config):
    """Create multiple AVDs based on the given configuration."""
    
    for i in range(config["numOfEnvs"]):
        avd_name = f"avd{i+1}"
        # Echo no to suppress the confirmation prompt
        cmd = f"echo no | avdmanager create avd -n {avd_name} -k \"system-images;android-34;google_apis;x86_64\""
        try:
            subprocess.run(cmd, shell=True, check=True)
            print(f"Successfully created AVD: {avd_name}")
        except subprocess.CalledProcessError as e:
            print(f"Failed to create AVD {avd_name}: {str(e)}")
            # raise RuntimeError(f"Failed to create AVD {avd_name}: {str(e)}") from e

def delete_avds(config):
    """Delete multiple AVDs based on the given configuration."""
    
    for i in range(config["numOfEnvs"]):
        avd_name = f"avd{i+1}"
        cmd = f"avdmanager delete avd -n {avd_name}"
        try:
            subprocess.run(cmd, shell=True, check=True)
            print(f"Successfully deleted AVD: {avd_name}")
        except subprocess.CalledProcessError as e:
            print(f"Failed to delete AVD {avd_name}: {str(e)}")
            raise RuntimeError(f"Failed to delete AVD {avd_name}: {str(e)}") from e

def start_emulators(config):
    """Start multiple emulators with the given configuration."""

    envs = []
    for i in range(config["numOfEnvs"]):
        # Create the task configuration using the path to the task definition file.
        task_config = config_classes.FilesystemTaskConfig(
            path='/home/coder/workspace/AppTestingEnvironment/sampleApps/mdp_0003.textproto'
        )

        # Create the emulator configuration including the avd_name.
        emulator_config = config_classes.EmulatorConfig(
            emulator_launcher=config_classes.EmulatorLauncherConfig(
                android_avd_home='/home/coder/.android/avd/',
                android_sdk_root='/opt/android-sdk',
                emulator_path='/opt/android-sdk/emulator/emulator',
                avd_name=f'avd{i+1}',  # This AVD name should be verified to exist.
                run_headless = config["run_headless"]
            ),
            adb_controller=config_classes.AdbControllerConfig(
                adb_path='/opt/android-sdk/platform-tools/adb'
            )
        )
        
        coordinator_config = config_classes.CoordinatorConfig(
            show_touches=False,
            show_pointer_location=False,
            # show_status_bar=False,
            # show_navigation_bar=False
        )

        # Check if AVD exists before loading the environment
        if not check_avd_exists(f'avd{i+1}'):
            raise ValueError(f"Specified AVD 'avd{i+1}' does not exist. Please create it or specify a different AVD.")

        # Create the AndroidEnv configuration object with the task and simulator config.
        android_env_config = config_classes.AndroidEnvConfig(
            task=task_config,
            simulator=emulator_config,
            coordinator=coordinator_config
        )

        # Now, load the environment using the config object.
        print("Loading the Android environment...")
        env = loader.load(config=android_env_config)
        print("Environment loaded successfully!")
        envs.append(env)
    return envs

def run_emulator_command(avd_name, log_file="emulator_logs.txt"):
    """
    Starts the Android Emulator with the specified AVD and redirects logs to a file.
    
    Args:
        avd_name (str): Name of the AVD to start.
        log_file (str): File to which emulator logs are redirected.
    """
    
    emulator_path = os.path.join(
        os.getenv('ANDROID_SDK_ROOT', '/opt/android-sdk'),
        'emulator',
        'emulator'
    )
    emulator_command = f"{emulator_path} -avd {avd_name} -grpc-use-token false"
    
    # Redirect logs to a file
    with open(log_file, "w") as log:
        subprocess.Popen(emulator_command, shell=True, stdout=log, stderr=log)
    print(f"Emulator started with AVD '{avd_name}'. Logs redirected to {log_file}.")

    
    # Redirect logs to a file
    with open(log_file, "w") as log:
        subprocess.Popen(emulator_command, shell=True, stdout=log, stderr=log)
    print(f"Emulator started with AVD '{avd_name}'. Logs redirected to {log_file}.")


def start_emulator():
    """Starts the Android environment and optionally launches the emulator if not already running."""
    
    # Create the task configuration using the path to the task definition file.
    task_config = config_classes.FilesystemTaskConfig(
        path='/home/coder/workspace/AppTestingEnvironment/sampleApps/mdp_0003.textproto'
    )

    # Create the emulator configuration including the avd_name.
    emulator_config = config_classes.EmulatorConfig(
        emulator_launcher=config_classes.EmulatorLauncherConfig(
            android_avd_home='/home/coder/.android/avd/',
            android_sdk_root='/opt/android-sdk',
            emulator_path='/opt/android-sdk/emulator/emulator',
            avd_name='avd1',  # This AVD name should be verified to exist.
            run_headless=False  # Set to True if you want the emulator to run without GUI
        ),
        adb_controller=config_classes.AdbControllerConfig(
            adb_path='/opt/android-sdk/platform-tools/adb'
        )
    )
    coordinator_config = config_classes.CoordinatorConfig(
        show_touches=False,
        show_pointer_location=False,
        show_status_bar=False,
        show_navigation_bar=False,
    )

    # Check if AVD exists before starting the emulator
    if not check_avd_exists('avd1'):
        raise ValueError("Specified AVD 'avd1' does not exist. Please create it or specify a different AVD.")

    # Optionally start the emulator
    run_emulator_command('avd1')

    # Create the AndroidEnv configuration object with the task and simulator config.
    android_env_config = config_classes.AndroidEnvConfig(
        task=task_config,
        simulator=emulator_config,
        coordinator=coordinator_config
    )

    # Load the Android environment
    print("Loading the Android environment...")
    env = loader.load(config=android_env_config)
    print("Environment loaded successfully!")
    return env


def init_wandb_get_key() -> str:
    """Retrieve the W&B API key from environment variables."""

    api_key = os.getenv('WANDB_API_KEY')
    return api_key

# Global variables to cache models
_pretrained_yolo_model = None
_base_yolo_model = None

def load_yolo_model(pretrained=True, copy_model=False):
    """Load YOLO model with specified configuration."""

    global _pretrained_yolo_model, _base_yolo_model
    
    try:
        if pretrained:
            if _pretrained_yolo_model is None or copy_model:
                current_dir = os.path.dirname(os.path.abspath(__file__))
                model_path = os.path.join(current_dir, "pretrained_weights_interactive_elements.pt")
                print(f"Loading YOLO model from: {model_path}")
                model = torch.hub.load(
                    "ultralytics/yolov5",
                    "custom",
                    path=model_path,
                    force_reload=True
                )
                if not copy_model:
                    _pretrained_yolo_model = model
            else:
                model = _pretrained_yolo_model
        else:
            if _base_yolo_model is None or copy_model:
                model = torch.hub.load(
                    "ultralytics/yolov5",
                    "yolov5s",
                    pretrained=False
                )
                if not copy_model:
                    _base_yolo_model = model
            else:
                model = _base_yolo_model
        
        return model
        
    except Exception as e:
        print(f"Error loading YOLO model: {e}")
        raise

def apply_wrappers(env, config):
    """Apply multiple wrappers to the given environment based on the configuration."""
    
    from android_env.wrappers import tap_action_wrapper
    from wrapper import pixels_only_wrapper
    from wrapper.gymnasium_interface_wrapper import GymnasiumInterfaceWrapper
    from wrapper.curiosity_wrapper import CuriosityWrapper
    from wrapper.element_detection_reward_wrapper import ElementDetectionRewardWrapper
    from wrapper.normalize_touch_action_wrapper import NormalizeTouchActionWrapper
    
    env = tap_action_wrapper.TapActionWrapper(env)

    # env = NormalizeTouchActionWrapper(env, config)

    env = pixels_only_wrapper.PixelsOnlyWrapper(env)
    
    env = GymnasiumInterfaceWrapper(env)  # DM -> Gymnasium conversion
    
    if config.get("use_curiosity", False):
        env = CuriosityWrapper(env, config)
        
    if config["use_element_detection_wrapper"]:
        env = ElementDetectionRewardWrapper(env, config, load_yolo_model(), device='cuda')

    return env

def do_config(do_wandb, api_key, config):
    """
    Initialize W&B with all configuration parameters.
    
    Args:
        do_wandb (bool): Whether to use W&B logging.
        api_key (str): W&B API key.
        config (dict): Full configuration dictionary from main.py.
    """

    if do_wandb:
        wandb.login(key=api_key)
        
        # Create a descriptive run name
        run_name = "curiosity_" if config['use_curiosity'] else "multiplied_curiosity_" if config['use_multiplied_curiosity'] else "score_"
        run_name += f"{config['policy_type'] if config['feature_extractor_class'] is None else config['feature_extractor_class'].__name__}_"
        run_name += f"{config['numOfEnvs']} env(s)_"
        run_name += ("pretrained" if config['use_pretrained'] else "scratch") if config['feature_extractor_class'] else ""
        run_name += ("_trainable" if config['train_feature_extractor'] else "_frozen") if config['feature_extractor_class'] else ""
        run_name += ("_element_detection" if config['use_element_detection_wrapper'] else "_no_element_detection")
        wandb_config = config.copy()
        wandb_config.pop('feature_extractor_class', None)
        wandb_config.pop('progress_callback', None)
        
        run = wandb.init(
            project="android_env_project",
            name=run_name,
            config=wandb_config,
            sync_tensorboard=True,
            monitor_gym=True,
            tags=[  # Optional: add tags for better organization
                config['policy_type'],
                'pretrained' if config['use_pretrained'] else 'from_scratch',
                'trainable_fe' if config['train_feature_extractor'] else 'frozen_fe',
                config['reward_type']
            ]
        )
    else:
        run = SimpleNamespace()
        run.id = ""
    
    return run, config