import os
from body.start_training import init_wandb_get_key, do_config
from wandb.integration.sb3 import WandbCallback
from gymnasium.wrappers import FrameStack
import logging
from stable_baselines3 import PPO
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.evaluation import evaluate_policy
from wrapper.android_vec_env import AndroidVecEnv

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class PeriodicSaveCallback(BaseCallback):
    """Callback to save the model periodically during training."""
    
    def __init__(self, save_freq: int, save_path: str, verbose: int = 0):
        """
        Initializes the callback.
        
        Args:
            save_freq (int): Number of timesteps between saves.
            save_path (str): Directory to save the models.
            verbose (int): Verbosity level.
        """
        
        super(PeriodicSaveCallback, self).__init__(verbose)
        self.save_freq = save_freq
        self.save_path = save_path

    def _on_step(self) -> bool:
        """Saves the model at defined intervals."""

        if self.num_timesteps % self.save_freq == 0:
            save_path = os.path.join(self.save_path, f"model_step_{self.num_timesteps}.zip")
            self.model.save(save_path)
            logger.info(f"Model saved at step {self.num_timesteps} to {save_path}")
        return True

def validate_config(config: dict):
    """Validates the configuration dictionary for required parameters."""
    
    required_keys = ["do_wandb", "policy_type", "reward_type", "total_timesteps", "save_model"]
    missing_keys = [key for key in required_keys if key not in config]
    if missing_keys:
        raise ValueError(f"Missing required config keys: {missing_keys}")

def initialize_model(env, config, run_id):
    """Initializes or loads a PPO model with the specified configuration."""
    
    try:
        # Dynamically choose the policy
        policy = config["policy_type"]

        # Get feature extractor configuration
        fe_kwargs = {
            "pretrained": config.get("use_pretrained", True),
            "trainable": config.get("train_feature_extractor", False)
        }

        model = PPO(
            policy=policy,
            env=env,
            learning_rate=config["learning_rate"],
            n_steps=config["n_steps"],
            tensorboard_log=f"./ppo_logs/{run_id}/",
            verbose=1,
            policy_kwargs={
                "normalize_images": False,
                **({"features_extractor_class": config["feature_extractor_class"],
                   "features_extractor_kwargs": fe_kwargs} if config["feature_extractor_class"] is not None else {})
            },
        )
        return model
    except Exception as e:
        logger.error(f"Error initializing model: {e}")
        raise


def train(config: dict):
    """Main function for training an RL model using PPO."""
    
    validate_config(config)

    # Handle W&B setup
    api_key = init_wandb_get_key() if config["do_wandb"] else ""
    run, run_config = do_config(config["do_wandb"], api_key, config)

    # Start the emulator and set up the environment
    env = AndroidVecEnv(config)

    # Create or load the model
    model = initialize_model(env, config, run.id)

    # Prepare for training
    if not config.get("use_without_training", False):
        logger.info("Starting training...")

        # Ensure the save directory exists
        save_path = config.get("save_path", "./models")
        os.makedirs(save_path, exist_ok=True)

        # Set up callbacks
        periodic_save_callback = PeriodicSaveCallback(save_freq=50000, save_path=save_path)
        callbacks = [periodic_save_callback]
        if config["do_wandb"]:
            callbacks.append(WandbCallback(gradient_save_freq=1000, verbose=2))
        
        if config["progress_callback"] is not None:
            callbacks.append(RLProgressCallback(
                progress_callback=config["progress_callback"],
                mode="train" if not config["use_without_training"] else "eval",
                verbose=1
            ))

        # Train the model
        model.learn(
            total_timesteps=config["total_timesteps"],
            callback=callbacks,
        )
        logger.info("Training completed.")

    # Save the final model if specified
    if config["save_model"]:
        final_model_path = os.path.join(save_path, f"{config['policy_type']}_final_model.zip")
        try:
            model.save(final_model_path)
            logger.info(f"Final model saved to {final_model_path}")
        except Exception as e:
            logger.error(f"Error saving final model: {e}")

    # Evaluate the model
    logger.info("Evaluating the trained model...")
    mean_reward, std_reward = evaluate_policy(model, env, n_eval_episodes=10)
    logger.info(f"Evaluation results - Mean reward: {mean_reward:.2f} +/- {std_reward:.2f}")

    # Clean up W&B run
    try:
        if run:
            run.finish()
            logger.info("W&B run finished and cleaned up.")
    except Exception as e:
        logger.warning(f"Error finishing W&B run: {e}")
