import numpy as np
from typing import List, Tuple, Dict, Union

class RewardCalculator:
    """
    Computes rewards based on image histogram comparisons and action sequences.

    Features:
    - Supports multiple distance metrics (Chi-squared, Bhattacharyya, Kullback-Leibler).
    - Maintains a memory of histograms for novelty detection.
    - Optionally uses a CNN model for embeddings.
    """
    
    def __init__(self, reward_type: str, use_cnn: bool = False) -> None:
        """
        Initializes the RewardCalculator.

        Args:
            reward_type (str): Distance metric for histogram comparison.
            use_cnn (bool): Whether to use CNN-based embeddings.
        """
        
        self.histogram_memory: List[np.ndarray] = []
        self.use_cnn = use_cnn
        self.reward_type = reward_type

        if use_cnn:
            from keras.applications import MobileNet
            self.embedding_model = MobileNet(weights='imagenet', include_top=False, pooling='avg', alpha=0.5)

    def reset(self) -> None:
        """Resets stored histograms."""

        self.histogram_memory = []

    def calculate_reward_temp(self, observation: np.ndarray, prev_observation: np.ndarray) -> float:
        """Computes a temporary reward based on observation differences."""

        if observation is None or prev_observation is None:
            return 0
        else:
            return self.calculate_reward_hist_mix(observation, prev_observation, distance_metric=self.reward_type)

    def calculate_reward_hist_mix(self, observation, prev_observation, distance_metric):
        """Calculates histogram-based reward with memory comparison."""

        # Use correct value if using accumulated reward 
        # Get histograms
        observation_hist, new = self.calculate_histogram(observation)
        prev_observation_hist, should_be_false = self.calculate_histogram(prev_observation)
        # This in case of stble baselines = False TODO: make flexbile
        # observation_hist, new = self.calculate_histogram(observation[0])
        # prev_observation_hist, should_be_false = self.calculate_histogram(prev_observation[0])
        current_distance = self.calculate_distance(observation_hist, prev_observation_hist, distance_metric)
        reward = current_distance
        
        # Calculate the distance between current and previous observation histograms
        current_distance = self.calculate_distance(observation_hist, prev_observation_hist, distance_metric)

        # Initialize variable to accumulate distances from memory comparison
        memory_distance_min = current_distance

        # Compare with all histograms in memory
        for hist in self.histogram_memory:
            memory_distance = self.calculate_distance(observation_hist, hist, distance_metric)
            if memory_distance < memory_distance_min:
                memory_distance_min = memory_distance
        return reward

    def calculate_reward(self, observation: np.ndarray, prev_observation: np.ndarray, 
                         bad_state: bool, reset_after: int, distance_metric: str, 
                         action: Dict[str, Union[int, List[float]]], last_action: Dict[str, Union[int, List[float]]]) -> float:
        """Computes reward based on observation changes and actions."""

        observation_hist, new = self.calculate_histogram(observation)
        prev_observation_hist, _ = self.calculate_histogram(prev_observation)

        reward = self.calculate_action_reward(action, last_action, new)

        if bad_state:
            reward += 5

        if new:
            reward = 1
        else:
            reward = 0

        return reward

    def calculate_action_reward(self, action: Dict[str, Union[int, List[float]]], 
                                last_action: Dict[str, Union[int, List[float]]], new: bool) -> float:
        """Evaluates action-based rewards and penalties."""

        if new:
            base_reward = 1
        else:
            base_reward = 0

        touch_distance = self.compute_touch_distance(action, last_action)

        if last_action['action_type'] == action['action_type']:
            if new:
                return 0.8
            return -0.2

        if (last_action['action_type'] in [0, 1] and action['action_type'] in [0, 1]) or \
           (last_action['action_type'] == 2 and action['action_type'] == 2):
            if new:
                return 1
            return 0.2

        return base_reward

    def compute_touch_distance(self, action1: Dict[str, Union[int, List[float]]], 
                               action2: Dict[str, Union[int, List[float]]]) -> float:
        """Computes the Euclidean distance between two touch positions."""

        x_diff = action1['touch_position'][0] - action2['touch_position'][0]
        y_diff = action1['touch_position'][1] - action2['touch_position'][1]
        return np.sqrt(x_diff**2 + y_diff**2)

    def calculate_histogram(self, observation: np.ndarray, bins: int = 30) -> Tuple[np.ndarray, bool]:
        """Generates a histogram for an image and checks if it's new."""

        histograms = []
        for i in range(3):  # 3 channel image
            channel = observation[i, :, :]
            channel_histogram, _ = np.histogram(channel, bins=bins, range=(0, 255))
            normalized_histogram = channel_histogram / channel_histogram.sum()
            histograms.append(normalized_histogram)
        
        histograms = np.concatenate(histograms)

        new = not any(np.array_equal(hist, histograms) for hist in self.histogram_memory)
        if new:
            self.histogram_memory.append(histograms)

        return histograms, new

    def calculate_distance(self, hist1: np.ndarray, hist2: np.ndarray, distance_metric: str) -> float:
        """Computes the distance between two histograms using the specified metric."""

        if distance_metric == "chi2":
            return np.sum((hist1 - hist2) ** 2 / (hist1 + hist2 + 1e-10))
        elif distance_metric == "bhattacharyya":
            return -np.log(np.sum(np.sqrt(hist1 * hist2)) + 1e-10)
        elif distance_metric == "kullback_leibler":
            return np.sum(hist1 * np.log((hist1 + 1e-10) / (hist2 + 1e-10)))
        else:
            raise ValueError("Invalid distance_metric. Use 'chi2', 'bhattacharyya', or 'kullback_leibler'.")

    def get_embedding_from_array(self, image_array: np.ndarray) -> np.ndarray:
        """Extracts a CNN-based embedding from an image array."""

        if not self.use_cnn:
            raise ValueError("CNN model is not initialized. Set use_cnn=True in the constructor.")

        from keras.applications.mobilenet import preprocess_input
        from keras.preprocessing.image import img_to_array

        image_array = np.resize(image_array, (224, 224, 3))
        image_array = np.clip(image_array, 0, 255).astype('uint8')
        image_array = preprocess_input(img_to_array(image_array))
        image_array = np.expand_dims(image_array, axis=0)
        
        return self.embedding_model.predict(image_array)