# main.py

import argparse
from body import training
from networks.custom_feature_extractor import CustomFeatureExtractor
import absl.logging
absl.logging.set_verbosity('error')  # Suppress info logs, show only errors

def parse_args():
    parser = argparse.ArgumentParser(description='Reinforcement Learning Training for Android Testing')
    
    # Training parameters
    parser.add_argument('--policy_type', type=str, default='MlpPolicy',
                      help='Policy type to use for training')
    parser.add_argument('--total_timesteps', type=int, default=1000000,
                      help='Total timesteps for training')
    parser.add_argument('--num_envs', type=int, default=1,
                      help='Number of environments to run in parallel')
    parser.add_argument('--run_headless', action='store_true', default=False,
                      help='Whether to run the emulator in headless mode')
    parser.add_argument('--learning_rate', type=float, default=3e-4,
                      help='Learning rate for training')
    parser.add_argument('--n_steps', type=int, default=2048,
                      help='Number of steps for training')
    
    # Environment configuration
    parser.add_argument('--env_name', type=str, default='reinforced_testing',
                      help='Name of the environment')
    parser.add_argument('--reward_type', type=str, default='bhattacharyya',
                      help='Type of reward function to use')
    
    # Curiosity configuration
    parser.add_argument('--use_curiosity', action='store_true', default=False,
                      help='Whether to use curiosity')
    parser.add_argument('--use_multiplied_curiosity', action='store_true', default=False,
                      help='Whether to use multiplied curiosity')
    parser.add_argument('--use_curiosity_reward_despite_termination', action='store_true', default=False,
                      help='Whether to use curiosity reward despite termination')
    
    # Element detection wrapper configuration
    parser.add_argument('--no_element_detection_wrapper', action='store_true', default=False,
                      help='Whether to use element detection wrapper')
    
    # Feature extractor configuration
    parser.add_argument('--no_custom_feature_extractor', action='store_true',
                      default=False,
                      help='Whether to use custom feature extractor')
    parser.add_argument('--no_pretrained', action='store_true',
                      default=False,
                      help='Whether to use pretrained YOLO weights')
    parser.add_argument('--train_feature_extractor', action='store_true',
                      default=False,
                      help='Whether to train the feature extractor')
    
    # Logging and saving
    parser.add_argument('--do_wandb', action='store_true',
                      help='Enable Weights & Biases logging')
    parser.add_argument('--save_model', action='store_true', default=True,
                      help='Save the trained model')
    parser.add_argument('--load_model_path', type=str, default=None,
                      help='Path to load a pretrained model')
    
    # Other options
    parser.add_argument('--use_without_training', action='store_true',
                      help='Use model without training')
    
    args = parser.parse_args()
    print(args)
    return args

def get_config(args):
    """Convert command line arguments to configuration dictionary."""
    config = {
        # Training parameters
        "policy_type": args.policy_type,
        "total_timesteps": args.total_timesteps,
        "numOfEnvs": args.num_envs,
        "run_headless": args.run_headless,
        "learning_rate": args.learning_rate,
        "n_steps": args.n_steps,
        
        # Environment configuration
        "env_name": args.env_name,
        "reward_type": args.reward_type,
        
        # Curiosity configuration
        "use_curiosity": args.use_curiosity,
        "use_multiplied_curiosity": args.use_multiplied_curiosity,
        "use_curiosity_reward_despite_termination": args.use_curiosity_reward_despite_termination,
        
        # Element detection wrapper configuration
        "use_element_detection_wrapper": not args.no_element_detection_wrapper,
        
        # Feature extractor configuration
        "feature_extractor_class": CustomFeatureExtractor if not args.no_custom_feature_extractor else None,
        "use_pretrained": not args.no_pretrained,
        "train_feature_extractor": args.train_feature_extractor,
        
        # Logging and saving
        "do_wandb": args.do_wandb,
        "do_stable_baselines": True,
        "save_model": args.save_model,
        "load_model_path": args.load_model_path,
        
        # Other options
        "use_without_training": args.use_without_training,
        "progress_callback": None,
    }
    return config

def main():
    # Parse command line arguments
    args = parse_args()
    
    # Get configuration with defaults and any command line overrides
    run_config = get_config(args)
    
    # Start training
    training.train(run_config)

if __name__ == "__main__":
    main()
