import torch
import torch.nn as nn
from stable_baselines3.common.torch_layers import BaseFeaturesExtractor
from gymnasium import spaces
from body.start_training import load_yolo_model

class YoloFeatureExtractor(nn.Module):
    def __init__(self, yolo_model=None, num_features_out=256, pretrained=True, trainable=False):
        super(YoloFeatureExtractor, self).__init__()
        
        # Constants for the YOLO model structure
        self.num_kernels_out_m1 = 128
        self.num_kernels_out_m2 = 256
        self.num_kernels_out_m3 = 512

        num_feature_map_m1 = 80
        num_feature_map_m2 = 40
        num_feature_map_m3 = 20

        num_reducer1_out = 1
        num_reducer2_out = 1
        num_reducer3_out = 1

        # Extract the first 9 modules from the YOLO model as feature extractor
        modules = list(yolo_model.children())[:9]

        # Create sequential models for each part of the YOLO model
        self.m1 = nn.Sequential(*modules[:5])
        self.m2 = nn.Sequential(*modules[5:7])
        self.m3 = nn.Sequential(*modules[7:])

        # Set trainable state for YOLO layers
        for module in [self.m1, self.m2, self.m3]:
            for param in module.parameters():
                param.requires_grad = trainable

        # Create parameter reducers (1x1 convolutions) - these must be trainable
        # This ensures that the final linear layer is not too large
        self.param_reducer1 = nn.Conv2d(self.num_kernels_out_m1, num_reducer1_out, kernel_size=1)
        self.param_reducer2 = nn.Conv2d(self.num_kernels_out_m2, num_reducer2_out, kernel_size=1)
        self.param_reducer3 = nn.Conv2d(self.num_kernels_out_m3, num_reducer3_out, kernel_size=1)
        
        # Final linear layer to produce the desired number of features - this must be trainable
        self.linear = nn.Sequential(
            nn.Linear(
                num_reducer1_out*num_feature_map_m1**2 +
                num_reducer2_out*num_feature_map_m2**2 +
                num_reducer3_out*num_feature_map_m3**2,
                num_features_out
            ), 
            nn.ReLU()
        )

    def forward(self, x):
        # Pad the input to match YOLO's expected input size
        x = nn.functional.pad(x, (160, 160, 0, 0), mode='constant', value=0)
        
        # Pass through the YOLO model parts
        x = self.m1(x)
        x2 = self.m2(x)
        x3 = self.m3(x2)

        # Apply parameter reducers
        x = self.param_reducer1(x)
        x2 = self.param_reducer2(x2)
        x3 = self.param_reducer3(x3)

        # Flatten the outputs
        x = torch.flatten(x, 1)
        x2 = torch.flatten(x2, 1)
        x3 = torch.flatten(x3, 1)

        # Concatenate and pass through final linear layer
        x = torch.cat((x, x2, x3), 1)
        x = self.linear(x)

        return x

class CustomFeatureExtractor(BaseFeaturesExtractor):
    """
    :param observation_space: (gym.Space)
    :param features_dim: (int) Number of features extracted.
    :param pretrained: (bool) Whether to use pretrained YOLO weights
    :param trainable: (bool) Whether to allow training of the feature extractor
    """

    def __init__(self, observation_space: spaces.Box, features_dim: int = 256, 
                 pretrained: bool = True, trainable: bool = False):
        super(CustomFeatureExtractor, self).__init__(observation_space, features_dim)
        
        # If trainable, we need a copy to prevent affecting other instances
        yolo_model = load_yolo_model(pretrained=pretrained, copy_model=trainable).model.model.model
        self.yolo_extractor = YoloFeatureExtractor(
            yolo_model=yolo_model,
            num_features_out=features_dim,
            pretrained=pretrained,
            trainable=trainable
        )

    def forward(self, observations: torch.Tensor) -> torch.Tensor:
        # Ensure input is on the correct device
        device = next(self.parameters()).device
        observations = observations.to(device)

        # Normalize input
        observations = observations.float() / 255.0

        # Extract features using YOLO
        features = self.yolo_extractor(observations)
        
        return features
    