import multiprocessing as mp
from typing import List, Optional, Tuple, Dict, Any, Callable
import numpy as np
from stable_baselines3.common.vec_env import SubprocVecEnv, VecEnv
import gymnasium as gym

def _worker(
    remote: mp.connection.Connection,
    parent_remote: mp.connection.Connection,
    env_fn: Callable
) -> None:
    """
    Worker function that runs in a separate process.
    
    Args:
        remote: Pipe connection to receive commands
        parent_remote: Pipe connection to send results
        env_fn: Function that creates the environment
    """
    parent_remote.close()
    env = env_fn()
    
    while True:
        try:
            cmd, data = remote.recv()
            if cmd == "step":
                observation, reward, terminated, truncated, info = env.step(data)
                remote.send(("step", (observation, reward, terminated, truncated, info)))
            elif cmd == "reset":
                observation, info = env.reset()
                remote.send(("reset", (observation, info)))
            elif cmd == "close":
                env.close()
                remote.close()
                break
            else:
                raise NotImplementedError(f"Unknown command {cmd}")
        except EOFError:
            break

class AndroidSubprocVecEnv(SubprocVecEnv):
    """
    A VecEnv that runs multiple Android environments in subprocesses.
    """
    
    def __init__(
        self,
        env_fns: List[Callable[[], gym.Env]],
        start_method: Optional[str] = None
    ):
        """
        Args:
            env_fns: List of functions that create environments to run in parallel
            start_method: Method used to start subprocesses ('spawn', 'fork', 'forkserver')
        """
        self.waiting = False
        self.closed = False
        
        if start_method is None:
            # Fork is not safe on macOS, spawn is the safest option
            start_method = "spawn"
            
        ctx = mp.get_context(start_method)
        
        self.remotes, self.work_remotes = zip(*[ctx.Pipe() for _ in range(len(env_fns))])
        self.processes = []
        
        for work_remote, remote, env_fn in zip(self.work_remotes, self.remotes, env_fns):
            args = (work_remote, remote, env_fn)
            process = ctx.Process(target=_worker, args=args, daemon=True)
            process.start()
            self.processes.append(process)
            work_remote.close()
            
        # Initialize parent class after setting up processes
        super().__init__(env_fns, start_method)
    
    def step_async(self, actions: np.ndarray) -> None:
        """Queue up environment steps to be run in subprocesses."""
        for remote, action in zip(self.remotes, actions):
            remote.send(("step", action))
        self.waiting = True
        
    def step_wait(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[Dict]]:
        """Wait for queued steps to complete and return results."""
        results = [remote.recv() for remote in self.remotes]
        self.waiting = False
        
        # Unpack results
        obs, rews, dones, truncs, infos = zip(*[r[1] for r in results])
        
        return self._process_obs(obs), np.stack(rews), np.stack(dones), list(infos)
        
    def close(self) -> None:
        """Clean up resources and close subprocesses."""
        if self.closed:
            return
            
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
                
        for remote in self.remotes:
            remote.send(("close", None))
            
        for process in self.processes:
            process.join()
            
        self.closed = True 