from typing import List, Optional, Tuple, Dict, Any, Callable
import numpy as np
from stable_baselines3.common.vec_env import VecEnv
import gymnasium as gym
from body.start_training import apply_wrappers, create_avds, start_emulators

class AndroidVecEnv(VecEnv):
    """
    Vectorized environment wrapper for Android environments.
    Allows running multiple Android environments in parallel.
    """
    def __init__(self, config):
        create_avds(config)
        self.envs = start_emulators(config)
        self.num_envs = config["numOfEnvs"]
        
        for i in range(config["numOfEnvs"]):
            self.envs[i] = apply_wrappers(self.envs[i], config)
        
        example_env = self.envs[0]
        observation_space = example_env.observation_space
        action_space = example_env.action_space
        
        for i in range(1, config["numOfEnvs"]):
            current_obs_space = self.envs[i].observation_space
            current_act_space = self.envs[i].action_space
            
            if current_obs_space != observation_space:
                raise ValueError(f"Env {i} has different observation space: {current_obs_space} vs {observation_space}")
            if current_act_space != action_space:
                raise ValueError(f"Env {i} has different action space: {current_act_space} vs {action_space}")
        
        super().__init__(self.num_envs, observation_space, action_space)
        
        self.device = config["device"] if "device" in config else "gpu"
        self.actions = None
        self.reset_infos = {}

    def reset(self) -> np.ndarray:
        """
        Reset the environments.
        """
        self.reset_infos = {}
        obs = []
        for env_idx, env in enumerate(self.envs):
            ob, info = env.reset()
            obs.append(ob)
            self.reset_infos[env_idx] = info
        
        return self._process_obs(obs)

    def step_async(self, actions: np.ndarray) -> None:
        """
        Store actions for all environments.
        """
        self.actions = actions

    def step_wait(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[Dict]]:
        """
        Perform step in all environments.
        """
        results = [env.step(a) for env, a in zip(self.envs, self.actions)]
        obs, rews, dones, truncs, infos = zip(*results)
        
        for env_idx, (done, env) in enumerate(zip(dones, self.envs)):
            if done or truncs[env_idx]:
                obs_i, info_i = env.reset()
                obs = list(obs)
                obs[env_idx] = obs_i
                infos[env_idx]["terminal_observation"] = obs_i
        
        processed_obs = self._process_obs(obs)
        stacked_rews = np.stack(rews)
        stacked_dones = np.stack(dones)
        return (processed_obs, stacked_rews, stacked_dones, list(infos))

    def close(self) -> None:
        """
        Clean up resources.
        """
        for env in self.envs:
            env.close()

    def _process_obs(self, obs: List[Any]) -> np.ndarray:
        """
        Process the observations.
        """
        try:
            return np.stack(obs)
        except Exception as e:
            raise

    def env_method(
        self,
        method_name: str,
        *method_args,
        indices: Optional[List[int]] = None,
        **method_kwargs
    ) -> List[Any]:
        """
        Call instance methods of vectorized environments.
        """
        indices = self._get_indices(indices)
        return [getattr(env, method_name)(*method_args, **method_kwargs) for env in [self.envs[i] for i in indices]]

    def get_attr(
        self, 
        attr_name: str, 
        indices: Optional[List[int]] = None
    ) -> List[Any]:
        """
        Get attribute from vectorized environments.
        """
        indices = self._get_indices(indices)
        return [getattr(env, attr_name) for env in [self.envs[i] for i in indices]]

    def set_attr(
        self,
        attr_name: str,
        value: Any,
        indices: Optional[List[int]] = None
    ) -> None:
        """
        Set attribute for vectorized environments.
        """
        indices = self._get_indices(indices)
        for env in [self.envs[i] for i in indices]:
            setattr(env, attr_name, value)

    def env_is_wrapped(
        self, 
        wrapper_class: type, 
        indices: Optional[List[int]] = None
    ) -> List[bool]:
        """
        Check if environments are wrapped with given wrapper.
        """
        indices = self._get_indices(indices)
        return [isinstance(env, wrapper_class) for env in [self.envs[i] for i in indices]]

    def _get_indices(self, indices: Optional[List[int]] = None) -> List[int]:
        """
        Convert indices to list of indices.
        """
        if indices is None:
            indices = range(self.num_envs)
        return indices