import cv2
import numpy as np
import dm_env
from dm_env import specs
import gymnasium
from gymnasium import spaces
from body.observation_checker import ImageObservationChecker
import wandb
from typing import Any
import time
from .curiosity_wrapper import CuriosityWrapper

class GymnasiumInterfaceWrapper(gymnasium.Env):
    """AndroidEnv with Gymnasium interface and single-frame observations."""

    def __init__(self, env, config):
        super().__init__()
        self._env = env
        self.screen_width = 84
        self.screen_height = 84
        self.faiss_threshold = config["faiss_threshold"] if "faiss_threshold" in config else 0  # Distance threshold for FAISS
        self.use_gpu = config["use_gpu"] if "use_gpu" in config else True  # Use GPU for FAISS and model if available
        # self.action_space = self._spec_to_space(self._env.action_spec())
        self.action_space = spaces.Box(low=0, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self._spec_to_space(self._env.observation_spec())
        self.use_curiosity = config["use_curiosity"]
        self.use_multiplied_curiosity = config["use_multiplied_curiosity"]
        self.use_curiosity_reward_despite_termination = config["use_curiosity_reward_despite_termination"]
        self.reward_type = config["reward_type"]
        self.do_wandb = config["do_wandb"]
        self.run_headless = config["run_headless"]
        self.episode_length = 0
        self.reward_sum = 0.0
        self.score = 0.0
        self.avg_reward = 0.0
        self.num_success = 0
        self.num_fail = 0
        self.touch_running_average_x = 0.0
        self.touch_running_average_y = 0.0
        self.sum_reward_for_success = 0.0
        self.sum_reward_for_fail = 0.0
        self.local_step_counter = 0  # Initialize step counter
        self.global_step_counter = 0
        self.max_steps = 256  # Maximum steps per episode
        self.observation_checker = ImageObservationChecker(threshold=self.faiss_threshold, use_gpu=self.use_gpu)
        self._latest_observation = None

    def _spec_to_space(self, spec: specs.Array) -> spaces.Space:
        """Converts dm_env specs to Gymnasium spaces."""
        if isinstance(spec, list):
            return spaces.Tuple([self._spec_to_space(s) for s in spec])
        if isinstance(spec, dict):
            return spaces.Dict({name: self._spec_to_space(s) for name, s in spec.items()})
        if isinstance(spec, specs.DiscreteArray):
            return spaces.Box(shape=(), dtype=spec.dtype, low=0, high=spec.num_values - 1)
        if isinstance(spec, specs.BoundedArray):
            low = np.full(spec.shape, spec.minimum, dtype=spec.dtype)
            high = np.full(spec.shape, spec.maximum, dtype=spec.dtype)
            return spaces.Box(low=low, high=high, dtype=spec.dtype)
        if isinstance(spec, specs.Array):
            low = 0 if spec.dtype == np.uint8 else -np.inf
            high = 255 if spec.dtype == np.uint8 else np.inf
            return spaces.Box(shape=(spec.shape), dtype=spec.dtype, low=low, high=high)
        raise ValueError('Unknown type for specs: {}'.format(spec))
    
    def preprocess_observation(self, observation):
        """Preprocess observation to match the expected input format."""
        # Ensure the pixels are in channel-first format (C, H, W)
        if observation.shape[-1] == 3:  # Convert from HWC (default) to CHW
            observation = np.transpose(observation, (2, 0, 1))

        # Ensure the pixel values are in uint8 format
        if observation.dtype != np.uint8:
            observation = (observation * 255).astype(np.uint8)
        
        
        if isinstance(observation, np.ndarray) and observation.shape == (3, 640, 320):
            resized_frame = cv2.resize(
                observation.transpose(1, 2, 0), (self.screen_width, self.screen_height)
            )
            grayscale_frame = cv2.cvtColor(resized_frame, cv2.COLOR_RGB2GRAY)
            return grayscale_frame.astype(np.uint8)  # (84, 84)
        else:
            raise ValueError(f"[ERROR] Unexpected observation format: {type(observation)}, {observation.shape}")

    def reset(self, seed=None, options=None):
        # Log metrics for the completed episode before resetting
        if self.do_wandb and self.episode_length > 0:  # Log only after at least one episode
            wandb.log({
                "reward_sum": self.reward_sum,
                "episode_length": self.episode_length,
                "score": getattr(self, "score", 0),  # Ensure score exists
                "entropy": self._calculate_entropy() if hasattr(self, "_calculate_entropy") else None,
            })

        # Reset metrics
        self._latest_observation = None
        self.score = 0.0
        self.reward_sum = 0.0
        self.episode_length = 0
        self.local_step_counter = 0  # Reset step counter

        # Reset the environment and get the initial observation
        timestep = self._env.reset()
        raw_observation = timestep.observation

        # Display raw observation for debugging
        if not self.run_headless:
            if isinstance(raw_observation, np.ndarray):
                debug_image = raw_observation.copy()
                if debug_image.shape[-1] == 3:  # If image is in RGB format
                    debug_image = cv2.cvtColor(debug_image, cv2.COLOR_RGB2BGR)  # Convert to BGR for cv2
                cv2.imshow('Raw Observation', debug_image)
                cv2.waitKey(1)  # Display for a short time

        # Preprocess the observation for the observation checker
        faiss_observation = self.preprocess_observation(raw_observation)

        # Clear observation embeddings
        self.observation_checker.clear()

        # Add the first observation to the observation checker
        self.observation_checker.add_image(faiss_observation)

        # Debugging: Verify observation matches observation_space
        # assert self.observation_space.contains(faiss_observation), (
        #     f"[ERROR] Reset observation does not match observation space: {faiss_observation.shape}"
        # )

        return raw_observation, {"discount": timestep.discount}

    def step(self, action):
        """
        Take a step in the environment and calculate a curiosity-driven reward.
        """
        
        # Handle OrderedDict format
        if isinstance(action, dict) or hasattr(action, '_asdict'):
            # Use the touch_position directly from the OrderedDict
            touch_position = action['touch_position']
            action_type = action['action_type']
        else:
            touch_position = action
            action_type = 0

        # Execute the action in the environment
        timestep = self._env.step({
            'action_type': action_type,
            'touch_position': tuple(touch_position)  # Convert numpy array to tuple
        })
        
        raw_observation = timestep.observation

        # Display raw observation for debugging
        if not self.run_headless:
            if isinstance(raw_observation, np.ndarray):
                debug_image = raw_observation.copy()
                if debug_image.shape[-1] == 3:  # If image is in RGB format
                    debug_image = cv2.cvtColor(debug_image, cv2.COLOR_RGB2BGR)  # Convert to BGR for cv2
                cv2.imshow('Raw Observation', debug_image)
                cv2.waitKey(1)  # Display for a short time
        
        # Preprocess the observation for the observation checker
        faiss_observation = self.preprocess_observation(raw_observation)

        # Check if observation has been seen before
        already_seen, obs_distance = self.observation_checker.is_seen(faiss_observation)

        if self.use_curiosity or self.use_curiosity_reward_despite_termination:
            reward = obs_distance / 20.0  # Curiosity reward scaling
        elif self.use_multiplied_curiosity:
            reward = self.score * obs_distance / 20.0  # Curiosity reward scaling
        else:
            reward = 1 if timestep.reward > 0 else 0
        done = timestep.last()
        
        if done and not self.use_curiosity_reward_despite_termination:
            reward = 0.0

        # Accumulate rewards and update metrics
        if timestep.reward > 0:
            self.score += 1.0
            self.num_success += 1
            self.sum_reward_for_success += reward
        else:
            self.num_fail += 1
            self.sum_reward_for_fail += reward
        self.reward_sum += reward
        self.episode_length += 1
        self.local_step_counter += 1
        self.global_step_counter += 1
        self._latest_observation = raw_observation
        self.avg_reward = (self.avg_reward * (self.global_step_counter - 1) + reward) / self.global_step_counter
        
        self.touch_running_average_x = (self.touch_running_average_x * (self.global_step_counter - 1) + touch_position[0]) / self.global_step_counter
        self.touch_running_average_y = (self.touch_running_average_y * (self.global_step_counter - 1) + touch_position[1]) / self.global_step_counter
        
        # Log metrics to W&B if enabled
        if self.do_wandb:
            wandb.log({
                "reward": reward,
                "obs_distance": obs_distance if self.use_curiosity else None,
                "score": self.score,
                "reward_sum": self.reward_sum,
                "episode_length": self.episode_length,
                "local_step_counter": self.local_step_counter,  # Log step count
                "global_step_counter": self.global_step_counter,
                "avg_reward": self.avg_reward,
                "avg_reward_for_fail": self.sum_reward_for_fail / self.num_fail if self.num_fail > 0 else 0,
                "avg_reward_for_success": self.sum_reward_for_success / self.num_success if self.num_success > 0 else 0,
                "num_success": self.num_success,
                "num_fail": self.num_fail,
                "ratio_success_fail": self.num_success / self.num_fail if self.num_fail > 0 else 0,
                "ratio_success_total": self.num_success / (self.num_success + self.num_fail) if (self.num_success + self.num_fail) > 0 else 0,
                "touch_running_average_x": self.touch_running_average_x,
                "touch_running_average_y": self.touch_running_average_y,
            }, step=self.global_step_counter)

        # Check if the episode is done
        truncated = self.local_step_counter >= self.max_steps
        
        if not self.run_headless:
            if self.use_curiosity:
                print(f"[DEBUG] Observation distance: {obs_distance}, Done: {done}, Truncated: {truncated}")
            # time.sleep(2)

        return raw_observation, float(reward), done, truncated, {"discount": timestep.discount}

    def render(self, mode="rgb_array"):
        return self._env.render() if mode == "rgb_array" else super().render(mode)

    def close(self):
        self._env.close()
        print(f"[DEBUG] Environment closed.")

def create_env(env, config):
    """Create a wrapped environment with all necessary wrappers"""
    env = GymnasiumInterfaceWrapper(env, config)  # Basic DM -> Gymnasium conversion
    if config.get("use_curiosity", False):
        env = CuriosityWrapper(env, config)  # Add curiosity rewards if enabled
    return env
