import cv2
import numpy as np
import gymnasium
from gymnasium import spaces
from body.observation_checker import ImageObservationChecker
import wandb

class CuriosityWrapper(gymnasium.Wrapper):
    """
    A wrapper that adds curiosity-driven rewards based on observation novelty.
    Uses FAISS to compare observations and generate curiosity rewards.
    """

    def __init__(self, env, config):
        super().__init__(env)
        self.screen_width = 84
        self.screen_height = 84
        self.config = config

        # Metrics
        self.episode_length = 0
        self.reward_sum = 0.0
        self.score = 0.0
        self.num_success = 0
        self.num_fail = 0
        self.sum_reward_for_success = 0.0
        self.sum_reward_for_fail = 0.0
        self.global_step_counter = 0
        
        # Initialize observation checker
        self.observation_checker = ImageObservationChecker(
            threshold=self.config.get("faiss_threshold", 0), 
            use_gpu=self.config.get("use_gpu", True)
        )

    def preprocess_observation(self, observation):
        """Preprocess observation to match the expected input format."""
        # Ensure the pixels are in channel-first format (C, H, W)
        if observation.shape[-1] == 3:  # Convert from HWC (default) to CHW
            observation = np.transpose(observation, (2, 0, 1))

        # Ensure the pixel values are in uint8 format
        if observation.dtype != np.uint8:
            observation = (observation * 255).astype(np.uint8)
        
        if isinstance(observation, np.ndarray) and observation.shape == (3, 640, 320):
            resized_frame = cv2.resize(
                observation.transpose(1, 2, 0), 
                (self.screen_width, self.screen_height)
            )
            grayscale_frame = cv2.cvtColor(resized_frame, cv2.COLOR_RGB2GRAY)
            return grayscale_frame.astype(np.uint8)  # (84, 84)
        else:
            raise ValueError(f"[ERROR] Unexpected observation format: {type(observation)}, {observation.shape}")

    def reset(self, **kwargs):
        # Log metrics for the completed episode
        if self.config.get("do_wandb", False) and self.episode_length > 0:
            wandb.log({
                "reward_sum": self.reward_sum,
                "episode_length": self.episode_length,
                "score": self.score,
            })

        # Reset metrics
        self.score = 0.0
        self.reward_sum = 0.0
        self.episode_length = 0

        # Reset the environment
        observation, info = self.env.reset(**kwargs)

        # Display debug visualization
        if not self.config.get("run_headless", True):
            self._debug_visualize(observation)

        # Process observation for curiosity system
        faiss_observation = self.preprocess_observation(observation)
        self.observation_checker.clear()
        self.observation_checker.add_image(faiss_observation)

        return observation, info

    def step(self, action):
        observation, reward, terminated, truncated, info = self.env.step(action)
        
        # Process observation for curiosity system
        faiss_observation = self.preprocess_observation(observation)
        already_seen, obs_distance = self.observation_checker.is_seen(faiss_observation)

        # Calculate curiosity reward
        if self.config.get("use_curiosity", False) or self.config.get("use_curiosity_reward_despite_termination", False):
            curiosity_reward = obs_distance / 20.0
        elif self.config.get("use_multiplied_curiosity", False):
            curiosity_reward = self.score * obs_distance / 20.0
        else:
            curiosity_reward = 1 if reward > 0 else 0

        # Handle termination
        if terminated and not self.config.get("use_curiosity_reward_despite_termination", False):
            curiosity_reward = 0.0

        # Update metrics
        if reward > 0:
            self.score += 1.0
            self.num_success += 1
            self.sum_reward_for_success += curiosity_reward
        else:
            self.num_fail += 1
            self.sum_reward_for_fail += curiosity_reward

        self.reward_sum += curiosity_reward
        self.episode_length += 1
        self.global_step_counter += 1

        # Log metrics
        if self.config.get("do_wandb", False):
            self._log_metrics(curiosity_reward, obs_distance)

        # Debug visualization
        if not self.config.get("run_headless", True):
            self._debug_visualize(observation)

        return observation, curiosity_reward, terminated, truncated, info

    def _debug_visualize(self, observation):
        """Helper method for debug visualization"""
        if isinstance(observation, np.ndarray):
            debug_image = observation.copy()
            if debug_image.shape[-1] == 3:
                debug_image = cv2.cvtColor(debug_image, cv2.COLOR_RGB2BGR)
            cv2.imshow('Raw Observation', debug_image)
            cv2.waitKey(1)

    def _log_metrics(self, reward, obs_distance = None):
        """Helper method for logging metrics to wandb"""
        wandb.log({
            "reward": reward,
            "obs_distance": obs_distance if obs_distance is not None else 0,
            "score": self.score,
            "reward_sum": self.reward_sum,
            "episode_length": self.episode_length,
            "global_step_counter": self.global_step_counter,
            "avg_reward_for_fail": self.sum_reward_for_fail / self.num_fail if self.num_fail > 0 else 0,
            "avg_reward_for_success": self.sum_reward_for_success / self.num_success if self.num_success > 0 else 0,
            "num_success": self.num_success,
            "num_fail": self.num_fail,
            "ratio_success_fail": self.num_success / self.num_fail if self.num_fail > 0 else 0,
            "ratio_success_total": self.num_success / (self.num_success + self.num_fail) if (self.num_success + self.num_fail) > 0 else 0,
        }, step=self.global_step_counter) 