
import gymnasium as gym
import numpy as np
import torch
from typing import Tuple, Dict, Any, Optional
from enum import Enum
import wandb

class DropoffType(Enum):
    """
    Enum for the different dropoff types.
    """
    LINEAR = "linear"
    EXPONENTIAL = "exponential"
    STEP = "step"

class ElementDetectionRewardWrapper(gym.Wrapper):
    """
    Wrapper that adds a penalty based on the distance of the touch position to the closest
    with yolo detected element on the screen.
    """
    def __init__(
        self,
        env: gym.Env,
        config: Dict[str, Any],
        yolo_model: torch.nn.Module,
        device: str = 'cuda',
        num_episodes_active: int = 1000,
        dropoff_type: DropoffType = DropoffType.STEP,
        penalty_strength: float = 1.0,
        action_grid: Optional[Tuple[int, int]] = None
    ):
        super().__init__(env)
        self.env = env
        self.device = device
        self.action_grid = action_grid
        self.num_episodes_active = num_episodes_active
        self.dropoff_type = dropoff_type
        self.penalty_strength = penalty_strength
        self.current_episode = 0
        self.do_wandb = config["do_wandb"]
        self.config = config
        
        self.yolo_model = yolo_model.to(self.device)
        self.yolo_model.eval()

    def get_penalty_factor(self) -> float:
        """
        Calculate the penalty impact based on current episode and dropoff type.
        Between 0 and 1.
        """
        if self.current_episode >= self.num_episodes_active:
            return 0.0
        
        progress = self.current_episode / self.num_episodes_active
        
        if self.dropoff_type == DropoffType.STEP:
            return 1.0
        elif self.dropoff_type == DropoffType.LINEAR:
            return 1.0 - progress
        elif self.dropoff_type == DropoffType.EXPONENTIAL:
            return np.exp(-5 * progress)
        
        return 1.0

    def step(self, action: Any) -> Tuple[Any, float, bool, bool, Dict[str, Any]]:
        """
        Step function that adds a penalty based on the distance of the touch position to the closest
        with yolo detected element on the screen.
        """
        observation, reward, terminated, truncated, info = self.env.step(action)
        touch_position = self.get_touch_position(action)
        detected_elements = self.detect_elements(observation)
        closest_distance = self.calculate_closest_distance(touch_position, detected_elements)
        modified_reward = self.modify_reward(reward, closest_distance)

        # Add info to the info dictionary for debugging
        info['reward_detected_elements'] = detected_elements
        info['reward_closest_element_distance'] = closest_distance
        info['reward_penalty_element_distance'] = modified_reward - reward
        info['current_penalty_factor'] = self.get_penalty_factor()
        info['current_episode'] = self.current_episode
        
        # Log to wandb if enabled
        if self.config["do_wandb"]:
            wandb.log({
                "reward_detected_elements": detected_elements,
                "reward_closest_element_distance": closest_distance,
                "reward_penalty_element_distance": modified_reward - reward,
                "current_penalty_factor": self.get_penalty_factor(),
                "current_episode": self.current_episode
            }, commit=False)

        if terminated or truncated:
            self.current_episode += 1

        return observation, modified_reward, terminated, truncated, info

    def get_touch_position(self, action: Any) -> Tuple[int, int]:
        """
        Get the touch position from the action independent of the type of action space.
        """
        if isinstance(action, dict) or hasattr(action, '_asdict'):
            touch_position = action['touch_position']
            return touch_position[0], touch_position[1]
        
        if isinstance(self.env.action_space, gym.spaces.Discrete):
            x = (action % self.action_grid[1] + 0.5) / self.action_grid[0]
            y = (action // self.action_grid[1] + 0.5) / self.action_grid[1]
            return x, y
        
        return action[0], action[1]
        
    def detect_elements(self, observation: np.ndarray) -> list:
        """
        Detect interactive elements on the screen.
        """
        img = observation.astype(np.float32)

        with torch.no_grad(), torch.amp.autocast('cuda'):
            results = self.yolo_model.forward(img)

        detections = []
        for *xyxy, conf, cls in results.xyxy[0].cpu().numpy():
            x1 = xyxy[0] / self.observation_space.shape[1]
            y1 = xyxy[1] / self.observation_space.shape[0]
            x2 = xyxy[2] / self.observation_space.shape[1]
            y2 = xyxy[3] / self.observation_space.shape[0]
            detections.append({
                'bbox': [x1, y1, x2, y2],
                'confidence': conf.item(),
                'class': int(cls)
            })
        
        return detections

    def calculate_closest_distance(self, touch_position: Tuple[int, int], detected_elements: list) -> float:
        """
        Calculate the closest distance between the touch position and any of the detected elements.
        """
        if not detected_elements:
            return float('inf')
        
        distances = []
        for element in detected_elements:
            bbox = element['bbox']
            center_x = (bbox[0] + bbox[2]) / 2
            center_y = (bbox[1] + bbox[3]) / 2
            distance = np.sqrt((touch_position[0] - center_x)**2 + (touch_position[1] - center_y)**2)
            distances.append(distance)
        
        return min(distances)

    def modify_reward(self, original_reward: float, closest_distance: float) -> float:
        """
        Modify the reward based on the closest distance to an interactive element.
        If there are no interactive elements detected by yolo, the reward is not modified.
        """
        if closest_distance == float('inf'):
            return original_reward
        else:
            penalty_factor = self.get_penalty_factor()
            distance_penalty = closest_distance * self.penalty_strength * penalty_factor
            return original_reward - distance_penalty

    def reset(self, **kwargs) -> Tuple[Any, Dict[str, Any]]:
        """
        Reset the environment.
        """
        observation, info = self.env.reset(**kwargs)
        return observation, info
