from typing import Any

import dm_env
from dm_env import specs
import gymnasium
from gymnasium import spaces
import numpy as np

class GymnasiumInterfaceWrapper(gymnasium.Env):
    """
    AndroidEnv with Gymnasium interface converting dm_env to Gymnasium.
    """
    def __init__(self, env: dm_env.Environment):
        self._env = env
        self.spec = None
        # Action space is always identical: action_type, touch_position_x, touch_position_y
        self.action_space = spaces.Box(0, 1, shape=(3,), dtype=np.float32)
        # Observation space depends on config of other wrappers
        self.observation_space = self._spec_to_space(self._env.observation_spec())
        
    def _spec_to_space(self, spec: specs.Array) -> spaces.Space:
        """
        Helper function to convert dm_env specs to Gymnasium spaces.
        """
        if isinstance(spec, list):
            return spaces.Tuple([self._spec_to_space(s) for s in spec])
        if isinstance(spec, dict):
            return spaces.Dict({name: self._spec_to_space(s) for name, s in spec.items()})
        if isinstance(spec, specs.DiscreteArray):
            return spaces.Box(shape=(), dtype=spec.dtype, low=0, high=spec.num_values - 1)
        if isinstance(spec, specs.BoundedArray):
            low = np.full(spec.shape, spec.minimum, dtype=spec.dtype)
            high = np.full(spec.shape, spec.maximum, dtype=spec.dtype)
            return spaces.Box(low=low, high=high, dtype=spec.dtype)
        if isinstance(spec, specs.Array):
            low = 0 if spec.dtype == np.uint8 else -np.inf
            high = 255 if spec.dtype == np.uint8 else np.inf
            return spaces.Box(shape=(spec.shape), dtype=spec.dtype, low=low, high=high)
        raise ValueError('Unknown type for specs: {}'.format(spec))

    def reset(self, seed=None, options=None) -> tuple[np.ndarray, dict]:
        """
        Reset the environment.
        """
        timestep = self._env.reset()

        info = {'discount': timestep.discount}
        return timestep.observation, info

    def step(self, action: dict[str, int]) -> tuple[np.ndarray, dict]:
        """Take a step in the base environment."""
        action_type = 0
        timestep = self._env.step({'action_type': action_type, 'touch_position': (action[1], action[2])})

        done = timestep.step_type == dm_env.StepType.LAST
        info = {'discount': timestep.discount}
        return timestep.observation, timestep.reward, done, False, info
    
    def render(self, mode="rgb_array"):
        """
        Render the environment.
        """
        return self._env.render() if mode == "rgb_array" else super().render()
    
    def close(self):
        """
        Close the environment.
        """
        self._env.close()
