from typing import Any

from android_env.wrappers import base_wrapper
from dm_env import specs
import dm_env

class PixelsOnlyWrapper(base_wrapper.BaseWrapper):
    """
    A wrapper class that removes all other observations and only keeps the pixel observations from the environment.

    Args:
        env (dm_env.Environment): The environment to wrap.

    Methods:
        __init__(env: dm_env.Environment): Initializes the NoActionsWrapper instance.
        _process_timestep(timestep: dm_env.TimeStep) -> dm_env.TimeStep: Processes a timestep and returns a modified timestep with only pixel observations.
        reset() -> dm_env.TimeStep: Resets the environment and returns the modified initial timestep with only pixel observations.
        step(action: int) -> dm_env.TimeStep: Takes a step in the environment with the given action and returns the modified timestep with only pixel observations.
    """

    def __init__(self, env: dm_env.Environment):
        """
        Initializes the NoActionsWrapper instance.

        Args:
            env (dm_env.Environment): The environment to wrap.
        """
        super().__init__(env)

    def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
        """
        Processes a timestep and returns a modified timestep with only pixel observations.

        Args:
            timestep (dm_env.TimeStep): The timestep to process.

        Returns:
            dm_env.TimeStep: The modified timestep with only pixel observations.
        """
        step_type, reward, discount, observation = timestep
        pixels = observation['pixels']
        return dm_env.TimeStep(
            step_type=step_type,
            reward=reward,
            discount=discount,
            observation=pixels)

    def reset(self) -> dm_env.TimeStep:
        """
        Resets the environment and returns the modified initial timestep with only pixel observations.

        Returns:
            dm_env.TimeStep: The modified initial timestep with only pixel observations.
        """
        timestep = self._env.reset()
        return self._process_timestep(timestep)

    def step(self, action: int) -> dm_env.TimeStep:
        """
        Takes a step in the environment with the given action and returns the modified timestep with only pixel observations.

        Args:
            action (int): The action to take in the environment.

        Returns:
            dm_env.TimeStep: The modified timestep with only pixel observations.
        """
        timestep = self._env.step(action)
        return self._process_timestep(timestep)

    def observation_spec(self) -> specs.Array | dict[str, specs.Array]:
        pixels_spec = self._env.observation_spec()['pixels']
        return pixels_spec