"""Wraps the AndroidEnv to expose an Gymnasium interface."""

from typing import Any

from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import gymnasium
from gymnasium import spaces
import numpy as np


class GymnasiumInterfaceWrapper(gymnasium.Env):
  """AndroidEnv with Gymnasium interface."""

  def __init__(self, env: dm_env.Environment):
    self._env = env
    self.spec = None
    # For StableBaselines3, action space cannot be a dict. Therefore, we flatten the dict to a box.
    # This results in a continuous value between 0 and 1 for the action type which must later be converted back to a discrete value.
    self.action_space = spaces.Box(0, 1, shape=(3,), dtype=np.float32)
    self.observation_space = self._spec_to_space(self._env.observation_spec())
    self.metadata = {'render.modes': ['rgb_array']}
    self._latest_observation = None

  def _spec_to_space(self, spec: specs.Array) -> spaces.Space:
    """Converts dm_env specs to Gymnasium spaces."""

    if isinstance(spec, list):
      return spaces.Tuple([self._spec_to_space(s) for s in spec])

    if isinstance(spec, dict):
      return spaces.Dict(
          {name: self._spec_to_space(s) for name, s in spec.items()}
      )

    if isinstance(spec, specs.DiscreteArray):
      return spaces.Box(
          shape=(),
          dtype=spec.dtype,
          low=0,
          high=spec.num_values-1)

    if isinstance(spec, specs.BoundedArray):
        low = np.full(spec.shape, spec.minimum, dtype=spec.dtype)
        high = np.full(spec.shape, spec.maximum, dtype=spec.dtype)
        return spaces.Box(low=low, high=high, dtype=spec.dtype)

    if isinstance(spec, specs.Array):
      if spec.dtype == np.uint8:
        low = 0
        high = 255
      else:
        low = -np.inf
        high = np.inf
      if(spec.shape == ()):
        return spaces.Box(shape=(1,), dtype=spec.dtype, low=low, high=high)
      return spaces.Box(shape=(spec.shape), dtype=spec.dtype, low=low, high=high)

    raise ValueError('Unknown type for specs: {}'.format(spec))

  def render(self, mode='rgb_array'):
    """Renders the environment."""
    if mode == 'rgb_array':
      if self._latest_observation is None:
        return

      return self._latest_observation['pixels']
    else:
      raise ValueError('Only supported render mode is rgb_array.')

  # Gynmasium needs a seed and options argument.
  def reset(self, seed=None, options=None) -> np.ndarray:
    self._latest_observation = None
    timestep = self._env.reset()
    # Not sure if this is the correct info. Was taken from step method.
    info = {'discount': timestep.discount}
    return (timestep.observation, info)

  def step(self, action: dict[str, int]) -> tuple[Any, ...]:
    """Take a step in the base environment."""
    # Convert continuous action_type to discrete action_type.
    if action[0] < 0.33:
      action_type = 0
    elif action[0] < 0.66:
      action_type = 1
    else:
      action_type = 2
    timestep = self._env.step({'action_type': action_type, 'touch_position': (action[1], action[2])})
    observation = timestep.observation
    self._latest_observation = observation
    reward = timestep.reward
    done = timestep.step_type == dm_env.StepType.LAST
    info = {'discount': timestep.discount}
    # Added False for truncated return value which is needed by gymnasium. Unsure if this is correct or incomplete.
    return observation, reward, done, False, info
